/*
 * Copyright (C) 2014-2019 Igor Maznitsa
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package com.igormaznitsa.zxpoly.components;

import static java.lang.Math.min;


import com.igormaznitsa.z80.Utils;
import com.igormaznitsa.z80.Z80;
import com.igormaznitsa.zxpoly.components.betadisk.BetaDiscInterface;
import com.igormaznitsa.zxpoly.utils.ConcurrentUByteArray;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;
import java.util.Locale;
import java.util.Random;
import java.util.logging.Level;
import java.util.logging.Logger;

@SuppressWarnings( {"unused", "FieldCanBeLocal", "NonAtomicOperationOnVolatileField"})
public final class Motherboard implements ZxPolyConstants {

  public static final long CPU_FREQ = 3540000L;
  public static final int TRIGGER_NONE = 0;
  public static final int TRIGGER_DIFF_MODULESTATES = 1;
  public static final int TRIGGER_DIFF_MEM_ADDR = 2;
  public static final int TRIGGER_DIFF_EXE_CODE = 4;
  private static final int NUMBER_OF_INT_BETWEEN_STATISTIC_UPDATE = 4;
  private static final Logger LOGGER = Logger.getLogger("MB");

  private static final int SPEC256_GFX_CORES = 8;

  private final ZxPolyModule[] modules;
  private final Z80[] spec256GfxCores;
  private final IoDevice[] ioDevices;
  private final IoDevice[] ioDevicesPreStep;
  private final IoDevice[] ioDevicesPostStep;
  private final ConcurrentUByteArray ram = new ConcurrentUByteArray(512 * 1024);
  private final VideoController video;
  private final KeyboardKempstonAndTapeIn keyboard;
  private final BetaDiscInterface betaDisk;
  private final float[] cpuLoad = new float[4];
  private volatile int port3D00 = (int) System.nanoTime() & 0xFF; // simulate noise after turning on
  private volatile boolean totalReset;
  private volatile int resetCounter;
  private int triggerMemAddress = 0;
  private int triggers = TRIGGER_NONE;
  private int intCounter;
  private volatile boolean videoFlashState;
  private boolean localResetForAllModules;
  private volatile BoardMode boardMode = BoardMode.ZXPOLY;
  private int statisticCounter = NUMBER_OF_INT_BETWEEN_STATISTIC_UPDATE;
  private volatile int gfxSyncRegsRecord = 0;
  private final Random rnd = new Random();

  public Motherboard(final RomData rom) {
    if (rom == null) {
      throw new NullPointerException("ROM must not be null");
    }
    this.modules = new ZxPolyModule[4];
    final List<IoDevice> iodevices = new ArrayList<>();
    for (int i = 0; i < this.modules.length; i++) {
      this.modules[i] = new ZxPolyModule(this, rom, i);
      iodevices.add(this.modules[i]);
    }

    this.betaDisk = new BetaDiscInterface(this);

    iodevices.add(this.betaDisk);

    this.keyboard = new KeyboardKempstonAndTapeIn(this);
    iodevices.add(keyboard);
    this.video = new VideoController(this);
    iodevices.add(video);
    iodevices.add(new KempstonMouse(this));
    this.ioDevices = iodevices.toArray(new IoDevice[0]);
    this.ioDevicesPreStep = Arrays.stream(this.ioDevices).filter(x -> (x.getNotificationFlags() & IoDevice.NOTIFICATION_PRESTEP) != 0).toArray(IoDevice[]::new);
    this.ioDevicesPostStep = Arrays.stream(this.ioDevices).filter(x -> (x.getNotificationFlags() & IoDevice.NOTIFICATION_POSTSTEP) != 0).toArray(IoDevice[]::new);

    // simulation of garbage in memory after power on
    for (int i = 0; i < this.ram.length(); i++) {
      this.ram.set(i, rnd.nextInt());
    }

    this.spec256GfxCores = new Z80[SPEC256_GFX_CORES];
    for (int i = 0; i < SPEC256_GFX_CORES; i++) {
      this.spec256GfxCores[i] = new Z80(this.modules[0].getCpu());
    }
  }

  public synchronized void forceResetAllCpu() {
    for (final ZxPolyModule p : this.modules) {
      p.getCpu().doReset();
    }
  }

  public void reset() {
    LOGGER.info("Full system reset");
    this.totalReset = true;
    this.resetCounter = 3;
  }

  public void resetAndRestoreRom(final RomData rom) {
    LOGGER.info("Restoring ROM");
    for (int i = 0; i < this.modules.length; i++) {
      this.modules[i].setRomData(rom);
    }
    LOGGER.info("Full system reset");
    this.totalReset = true;
    this.resetCounter = 3;
  }


  public boolean set3D00(final int value, final boolean force) {
    if (is3D00NotLocked() || force) {
      this.port3D00 = value;
      LOGGER.log(Level.INFO, "set #3D00 to " + Utils.toHex(value));

      if ((value & PORTw_ZXPOLY_RESET) != 0) {
        for (final ZxPolyModule m : this.modules) {
          m.prepareLocalReset();
        }
        this.localResetForAllModules = true;
      }

      this.video.setVideoMode((this.port3D00 >> 2) & 0x7);
      return true;
    } else {
      LOGGER.info("Rejected new value for #3D00 because it is locked");
      return false;
    }
  }

  public int findFirstDiffAddrInModuleMemory() {
    int addr = -1;
    for (int i = 0; i < 0x20000; i++) {
      final int m0 = this.modules[0].readHeap(i);
      final int m1 = this.modules[1].readHeap(i);
      final int m2 = this.modules[2].readHeap(i);
      final int m3 = this.modules[3].readHeap(i);

      final int summ = m0 + m1 + m2 + m3;
      if ((m0 << 2) != summ) {
        addr = i;
        break;
      }
    }
    return addr;
  }

  public float getCpuActivity(final int cpuIndex) {
    return this.cpuLoad[cpuIndex];
  }

  public BetaDiscInterface getBetaDiskInterface() {
    return this.betaDisk;
  }

  public int get3D00() {
    return this.port3D00;
  }

  public int getMappedCpuIndex() {
    return (this.port3D00 >>> 5) & 3;
  }

  public boolean is3D00NotLocked() {
    return (this.port3D00 & PORTw_ZXPOLY_BLOCK) == 0;
  }

  public boolean isSlaveModulesInWaitMode() {
    return (this.port3D00 & PORTw_ZXPOLY_nWAIT) == 0;
  }

  public boolean isFlashActive() {
    return this.videoFlashState;
  }

  public Z80 getMasterCpu() {
    return this.modules[0].getCpu();
  }

  public int getTriggers() {
    return this.triggers;
  }

  public void setTrigger(final int flag) {
    this.triggers |= flag;
  }

  public void resetTrigger(final int flag) {
    this.triggers &= ~flag;
  }

  public int getMemTriggerAddress() {
    return this.triggerMemAddress;
  }

  public void setMemTriggerAddress(final int address) {
    this.triggerMemAddress = address;
  }

  public void setGfxAlignRegisters(final String registers) {
    this.gfxSyncRegsRecord = Z80.parseAndPackRegAlignValue(registers);
    LOGGER.info("Set GFX register list for aligning, '" + registers + "' = " + Integer.toBinaryString(this.gfxSyncRegsRecord));
  }

  public int step(final boolean signalInt, final boolean processStep) {
    this.localResetForAllModules = false;

    final boolean resetStatisticsAtModules;

    int result = TRIGGER_NONE;

    if (signalInt) {
      this.statisticCounter--;
      if (this.statisticCounter <= 0) {
        for (int i = 0; i < 4; i++) {
          this.cpuLoad[i] = min(1.0f, (float) (this.modules[i].getActiveMCyclesBetweenInt() / NUMBER_OF_INT_BETWEEN_STATISTIC_UPDATE) / (float) (VideoController.CYCLES_BETWEEN_INT));
        }
        this.statisticCounter = NUMBER_OF_INT_BETWEEN_STATISTIC_UPDATE;
        resetStatisticsAtModules = true;
      } else {
        resetStatisticsAtModules = false;
      }

      this.intCounter++;
      if (this.intCounter >= 25) {
        this.intCounter = 0;
        this.videoFlashState = !this.videoFlashState;
      }
    } else {
      resetStatisticsAtModules = false;
    }

    if (processStep) {
      final ZxPolyModule[] modules = this.modules;

      final boolean signalReset = this.totalReset;
      this.totalReset = false;

      if (this.resetCounter > 0) {
        this.resetCounter--;
        if (this.resetCounter == 0) {
          this.set3D00(0, true);
        }
      }

      for (final IoDevice device : this.ioDevicesPreStep) {
        device.preStep(signalReset, signalInt);
      }

      final long initialMachineCycleCounter = modules[0].getCpu().getMachineCycles();

      final BoardMode mode = this.getBoardMode();

      if (mode == BoardMode.ZXPOLY) {

        final boolean zx0halt;
        final boolean zx1halt;
        final boolean zx2halt;
        final boolean zx3halt;

        switch ((int) System.nanoTime() & 0x3) {
          case 0: {
            zx0halt = modules[0].step(signalReset, signalInt, resetStatisticsAtModules);
            if (localResetForAllModules) {
              return result;
            }
            zx3halt = modules[3].step(signalReset, signalInt, resetStatisticsAtModules);
            zx2halt = modules[2].step(signalReset, signalInt, resetStatisticsAtModules);
            zx1halt = modules[1].step(signalReset, signalInt, resetStatisticsAtModules);
          }
          break;
          case 1: {
            zx1halt = modules[1].step(signalReset, signalInt, resetStatisticsAtModules);
            zx2halt = modules[2].step(signalReset, signalInt, resetStatisticsAtModules);
            zx0halt = modules[0].step(signalReset, signalInt, resetStatisticsAtModules);
            if (this.localResetForAllModules) {
              return result;
            }
            zx3halt = modules[3].step(signalReset, signalInt, resetStatisticsAtModules);
          }
          break;
          case 2: {
            zx3halt = modules[3].step(signalReset, signalInt, resetStatisticsAtModules);
            zx0halt = modules[0].step(signalReset, signalInt, resetStatisticsAtModules);
            if (this.localResetForAllModules) {
              return result;
            }
            zx1halt = modules[1].step(signalReset, signalInt, resetStatisticsAtModules);
            zx2halt = modules[2].step(signalReset, signalInt, resetStatisticsAtModules);
          }
          break;
          case 3: {
            zx2halt = modules[2].step(signalReset, signalInt, resetStatisticsAtModules);
            zx3halt = modules[3].step(signalReset, signalInt, resetStatisticsAtModules);
            zx1halt = modules[1].step(signalReset, signalInt, resetStatisticsAtModules);
            zx0halt = modules[0].step(signalReset, signalInt, resetStatisticsAtModules);
            if (this.localResetForAllModules) {
              return result;
            }
          }
          break;
          default:
            throw new Error("Unexpected value");
        }

        if (is3D00NotLocked() && (zx0halt || zx1halt || zx2halt || zx3halt)) {
          // a cpu has met halt and we need process notification
          if (zx0halt) {
            doModuleHaltNotification(0);
          }
          if (zx1halt) {
            doModuleHaltNotification(1);
          }
          if (zx2halt) {
            doModuleHaltNotification(2);
          }
          if (zx3halt) {
            doModuleHaltNotification(3);
          }
        }
      } else {
        // ZX128 and SPEC256 modes
        final ZxPolyModule masterModule = modules[0];
        if (this.boardMode == BoardMode.SPEC256) {
          final Z80 mainCpu = masterModule.getCpu();
          masterModule.saveInternalCopyForGfx();
          masterModule.step(signalReset, signalInt, resetStatisticsAtModules);
          for (int i = 0; i < SPEC256_GFX_CORES; i++) {
            masterModule.stepWithGfxCpu(i + 1, this.spec256GfxCores[i], signalReset, signalInt);
            this.spec256GfxCores[i].alignRegisterValuesWith(mainCpu, this.gfxSyncRegsRecord);
          }
        } else {
          masterModule.step(signalReset, signalInt, resetStatisticsAtModules);
        }
      }

      final long spentMachineCycles = modules[0].getCpu().getMachineCycles() - initialMachineCycleCounter;

      for (final IoDevice device : this.ioDevicesPostStep) {
        device.postStep(spentMachineCycles);
      }

      final int curTriggers = this.triggers;

      if (curTriggers != TRIGGER_NONE) {
        if ((curTriggers & TRIGGER_DIFF_MODULESTATES) != 0 && !haveModulesSamePositionAndMode()) {
          this.triggers = this.triggers & ~TRIGGER_DIFF_MODULESTATES;
          result |= TRIGGER_DIFF_MODULESTATES;
        }
        if ((curTriggers & TRIGGER_DIFF_MEM_ADDR) != 0) {
          int val = modules[0].readAddress(this.triggerMemAddress);
          for (int i = 1; i < 4; i++) {
            if (val != modules[i].readAddress(this.triggerMemAddress)) {
              result |= TRIGGER_DIFF_MEM_ADDR;
              this.triggers = this.triggers & ~TRIGGER_DIFF_MEM_ADDR;
              break;
            }
          }
        }

        if ((curTriggers & TRIGGER_DIFF_EXE_CODE) != 0) {
          final int m1ExeByte = modules[0].getCpu().getLastM1InstructionByte();
          final int exeByte = modules[0].getCpu().getLastInstructionByte();

          for (int i = 1; i < 4; i++) {
            if (m1ExeByte != modules[i].getCpu().getLastM1InstructionByte() || exeByte != modules[i].getCpu().getLastInstructionByte()) {
              result |= TRIGGER_DIFF_EXE_CODE;
              this.triggers = this.triggers & ~TRIGGER_DIFF_EXE_CODE;
              break;
            }
          }
        }
      }
    }
    return result;
  }

  public void syncSpec256GpuStates() {
    final Z80 mainCpu = this.modules[0].getCpu();
    for (final Z80 spec256GfxCore : this.spec256GfxCores) {
      spec256GfxCore.fillByState(mainCpu);
    }
  }

  private boolean haveModulesSamePositionAndMode() {
    final int pc = this.modules[0].getCpu().getRegister(Z80.REG_PC);
    final int sp = this.modules[0].getCpu().getRegister(Z80.REG_SP);
    final int im = this.modules[0].getCpu().getIM();
    final boolean iff1 = this.modules[0].getCpu().isIFF1();
    final boolean iff2 = this.modules[0].getCpu().isIFF2();

    boolean result = true;

    for (int i = 1; result && i < 4; i++) {
      result = pc == this.modules[i].getCpu().getRegister(Z80.REG_PC)
          && sp == this.modules[i].getCpu().getRegister(Z80.REG_SP)
          && im == this.modules[i].getCpu().getIM()
          && iff1 == this.modules[i].getCpu().isIFF1()
          && iff2 == this.modules[i].getCpu().isIFF2();

    }
    return result;
  }

  private void doModuleHaltNotification(final int moduleIndex) {
    final ZxPolyModule module = this.modules[moduleIndex];
    final int reg1 = module.getReg1WrittenData();
    final boolean sendInt = (reg1 & ZXPOLY_wREG1_HALT_NOTIFY_INT) != 0;
    final boolean sendNmi = (reg1 & ZXPOLY_wREG1_HALT_NOTIFY_NMI) != 0;
    if (sendInt) {
      if ((reg1 & ZXPOLY_wREG1_CPU0) != 0) {
        this.modules[0].prepareLocalInt();
      }
      if ((reg1 & ZXPOLY_wREG1_CPU1) != 0) {
        this.modules[1].prepareLocalInt();
      }
      if ((reg1 & ZXPOLY_wREG1_CPU2) != 0) {
        this.modules[2].prepareLocalInt();
      }
      if ((reg1 & ZXPOLY_wREG1_CPU3) != 0) {
        this.modules[3].prepareLocalInt();
      }
    }

    if (sendNmi) {
      if ((reg1 & ZXPOLY_wREG1_CPU0) != 0) {
        this.modules[0].prepareLocalNmi();
      }
      if ((reg1 & ZXPOLY_wREG1_CPU1) != 0) {
        this.modules[1].prepareLocalNmi();
      }
      if ((reg1 & ZXPOLY_wREG1_CPU2) != 0) {
        this.modules[2].prepareLocalNmi();
      }
      if ((reg1 & ZXPOLY_wREG1_CPU3) != 0) {
        this.modules[3].prepareLocalNmi();
      }
    }
  }

  public BoardMode getBoardMode() {
    return this.boardMode;
  }

  public void setBoardMode(final BoardMode newMode, final boolean doReset) {
    if (this.boardMode != newMode) {
      LOGGER.log(Level.INFO, "Motherboard mode changed to " + newMode);
      this.boardMode = newMode;
      if (doReset) {
        this.reset();
      }
    }
  }

  public ZxPolyModule[] getModules() {
    return this.modules;
  }

  public VideoController getVideoController() {
    return this.video;
  }

  public int readRam(final ZxPolyModule module, final int address) {
    return this.ram.get(address);
  }

  public void writeRam(final ZxPolyModule module, final int heapAddress, final int value) {
    this.ram.set(heapAddress, value);
  }

  public int readBusIo(final ZxPolyModule module, final int port) {
    final int mappedCPU = getMappedCpuIndex();
    int result = -1;

    if (this.getBoardMode() == BoardMode.ZXPOLY && (module.getModuleIndex() == 0 && mappedCPU > 0)) {
      final ZxPolyModule destmodule = modules[mappedCPU];
      result = this.ram.get(destmodule.ramOffset2HeapAddress(destmodule.read7FFD(), port));
      destmodule.prepareLocalInt();
    } else {
      IoDevice firstDetectedActiveDevice = null;
      for (final IoDevice device : this.ioDevices) {
        final int data = device.readIo(module, port);
        if (data < 0) {
          continue;
        }

        if (result < 0) {
          result = data;
          firstDetectedActiveDevice = device;
        } else {
          final int prevResult = result;
          result |= data;
          if (prevResult != result) {
            LOGGER.log(Level.WARNING, "Detected IO collision during read: " + firstDetectedActiveDevice + ", " + device.getName() + " port #" + Integer.toHexString(port).toUpperCase(Locale.ENGLISH));
          }
        }
      }

      if (result < 0) {
        // all IO devices in Z state, some simulation of "port FF"
        result = this.rnd.nextInt(100) > 90 ? 0xFF :
            this.modules[0].readVideo(0x1800 + rnd.nextInt(0x300));
      }
    }
    return result;
  }

  public void writeBusIo(final ZxPolyModule module, final int port, final int value) {
    final int mappedCpu = getMappedCpuIndex();
    final int moduleIndex = module.getModuleIndex();

    if (this.getBoardMode() == BoardMode.ZXPOLY) {
      if (moduleIndex == 0) {
        if (port == PORTrw_ZXPOLY) {
          set3D00(value, false);
        } else {
          if (mappedCpu > 0) {
            final ZxPolyModule destmodule = this.modules[mappedCpu];
            this.ram.set(destmodule.ramOffset2HeapAddress(destmodule.read7FFD(), port), value);
            destmodule.prepareLocalNmi();
          } else {
            for (final IoDevice d : this.ioDevices) {
              d.writeIo(module, port, value);
            }
          }
        }
      } else {
        for (final IoDevice d : this.ioDevices) {
          d.writeIo(module, port, value);
        }
      }
    } else {
      for (final IoDevice d : this.ioDevices) {
        d.writeIo(module, port, value);
      }
    }
  }

  public <T> T findIoDevice(final Class<T> klazz) {
    T result = null;
    for (final IoDevice d : this.ioDevices) {
      if (klazz.isInstance(d)) {
        result = klazz.cast(d);
        break;
      }
    }
    return result;
  }

  public synchronized void resetIoDevices() {
    final IoDevice[] devices = this.ioDevices;
    final int devicesNum = devices.length;
    for (int i = 0; i < devicesNum; i++) {
      devices[i].doReset();
    }
  }
}
